% Annulus - an annulus for ray tracing (see Shape)
%

classdef Annulus < Shape
  properties(GetAccess = public, SetAccess = private)
    length = 1;
    outerRadius = 1;
    innerRadius = 1;
  end
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Public Methods
  methods
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Constructor and Configuration
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function obj = Annulus(varargin)
      switch nargin
        case 0
          % do nothing
        case 1
          arg = varargin{1};
          if isscalar(arg) && isa(arg, 'Annulus')
            obj.setTo(arg);
          elseif Vec3.isSingleVector(arg(:))
            obj.setLengthAndRadii(arg(1), arg(2), arg(3));
          else
            error('Single argument must be a 3-vector or a Annulus.')
          end
        case 3
          arg1 = varargin{1};
          arg2 = varargin{2};
          arg3 = varargin{3};
          if isscalar(arg1) && isnumeric(arg1) && ...
              isscalar(arg2) && isnumeric(arg2) && ...
              isscalar(arg3) && isnumeric(arg3)
            obj.setLengthAndRadii(arg1, arg2, arg3);
          else
            error('Argument must be a numeric scalar.')
          end
        otherwise
          error('Constructor requires 0, 1 or 3 arguments.')
      end
    end
    
    function obj = copy(sourceObj)
      obj = Annulus(sourceObj);
    end
    
    function setTo(obj, sourceObj)
      Shape.setTo(obj, sourceObj);
      obj.length = sourceObj.length;
      obj.outerRadius = sourceObj.outerRadius;
      obj.innerRadius = sourceObj.innerRadius;
    end
    
    function setLengthAndRadii(obj, lenArg, oRadArg, iRadArg)
      obj.length = lenArg;
      obj.outerRadius = oRadArg;
      obj.innerRadius = iRadArg;
      obj.maxRadius = oRadArg;
    end
    
    function setTopPosition(obj, pos)
      pos = Vec3.returnSingleVector(pos) - [0; 0; obj.length / 2];
      obj.setPosition(pos);
    end
    
    function setBottomPosition(obj, pos)
      pos = Vec3.returnSingleVector(pos) + [0; 0; obj.length / 2];
      obj.setPosition(pos);
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Shape Properties
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function vol = volume(obj)
      vol = pi * obj.length * (obj.outerRadius^2 - obj.innerRadius^2);
    end
    function area = surfaceArea(obj)
      area = 2 * pi * ((obj.outerRadius^2 - obj.innerRadius^2) + ...
        obj.length * (obj.outerRadius + obj.innerRadius));
    end
    function area = crossSectionZ(obj)
      area = pi * (obj.outerRadius^2 - obj.innerRadius^2);
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Surface Properties
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function vNorm = surfaceNormal(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      
      [isOn, faceNumber] = isOnSurface(obj, pos, true);
      
      isOnOuter = bitand(faceNumber, 1) ~= 0;
      isOnInner = bitand(faceNumber, 2) ~= 0;
      isOnTop = bitand(faceNumber, 4) ~= 0;
      isOnBottom = bitand(faceNumber, 8) ~= 0;
      
      vNorm = pos;
      vNorm(1:2, isOnTop | isOnBottom) = 0;
      vNorm(3, isOnTop) = 1;
      vNorm(3, isOnBottom) = -1;
      
      vNorm(3, isOnOuter | isOnInner) = 0;
      vNorm(:, isOnOuter) = Vec3.normalize(vNorm(:, isOnOuter));
      vNorm(:, isOnInner) = Vec3.normalize(-vNorm(:, isOnInner));
      
      if ~isLocal
        vNorm = obj.mLocalToGlobal * vNorm;
      end
    end
    
    function cSurf = surfaceColor(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      halfLen = obj.length / 2;
      
      isOnTop = abs(halfLen - pos(3, :)) < Shape.surfaceTolerance;
      isOnBottom = abs(pos(3, :) + halfLen) < Shape.surfaceTolerance;
      isOnRound = ~(isOnTop | isOnBottom);
      
      cSurf = Vec3.copy(obj.surfColor, size(pos, 2));
      if any(isOnTop)
        cSurf(:, isOnTop) = Vec3.sum(cSurf(:, isOnTop) * 0.9, 0.1 * [0; 0; 1]);
      end
      if any(isOnBottom)
        cSurf(:, isOnBottom) = Vec3.sum(cSurf(:, isOnBottom) * 0.9, 0.1 * [0; 1; 0]);
      end
      if any(isOnRound)
        cSurf(:, isOnRound) = Vec3.sum(cSurf(:, isOnRound) * 0.9, 0.1 * [1; 0; 0]);
      end
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Inside and Outside
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function [isOn, faceNumber] = isOnSurface(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      posT = pos';
      halfLen = obj.length / 2;
      
      % is it on the outer cylinder?
      radXY = sum(posT(:, 1:2).^2, 2);
      isInCylZ = posT(:, 3) <= halfLen & posT(:, 3) >= -halfLen;
      isOnOuter = isInCylZ & ...
        abs(radXY - obj.outerRadius^2) < Shape.surfaceTolerance;
      
      % is it on the inner cylinder ?
      isOnInner = isInCylZ & ...
        abs(radXY - obj.innerRadius^2) < Shape.surfaceTolerance;
      
      % is it on the top?
      isOnTop = abs(posT(:, 3) - halfLen) < Shape.surfaceTolerance;
      
      % is it on the bottom?
      isOnBottom = abs(posT(:, 3) + halfLen) < Shape.surfaceTolerance;
      
      % is it on the surface?
      faceNumber = isOnOuter + 2 * isOnInner + 4 * isOnTop + 8 * isOnBottom;
      isOn = faceNumber ~= 0;
    end
    function isIn = isInside(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      posT = pos';
      
      % is it in the cylinder radius and between the ends?
      halfLen = obj.length / 2;
      radXY = sum(posT(:, 1:2).^2, 2);
      isIn = radXY < obj.outerRadius^2 & radXY > obj.innerRadius^2 & ...
        posT(:, 3) < halfLen & posT(:, 3) > -halfLen;
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Collision Checking
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function tHit = hitSurface(obj, pos, vel, isLocal)
      isLocal = nargin > 3 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        [pos, vel] = obj.transformTo(pos, vel);
      end
      halfLen = obj.length / 2;
      
      % make XY only vectors
      [posMatch, velMatch] = Vec3.matchSize(pos, vel);
      pXY = posMatch(1:2, :);
      vXY = velMatch(1:2, :);
      
      rXY2 = sum(pXY.^2)';
      
      %%%%%%%%%%%%%%%%%%% not done
      error('only hitFromOutside implemented')
    end
    
    function tHit = hitFromInside(obj, pos, vel, isLocal)
      isLocal = nargin > 3 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        [pos, vel] = obj.transformTo(pos, vel);
      end
      halfLen = obj.length / 2;
      
      % make XY only vectors
      [posMatch, velMatch] = Vec3.matchSize(pos, vel);
      pXY = posMatch(1:2, :);
      vXY = velMatch(1:2, :);
      
      rXY2 = sum(pXY.^2)';
      
      %%%%%%%%%%%%%%%%%%% not done
      error('only hitFromOutside implemented')
    end
    
    function tHit = hitFromOutside(obj, pos, vel, isLocal)
      isLocal = nargin > 3 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        [pos, vel] = obj.transformTo(pos, vel);
      end
      halfLen = obj.length / 2;
      
      % make XY only vectors
      [posMatch, velMatch] = Vec3.matchSize(pos, vel);
      pXY = posMatch(1:2, :);
      vXY = velMatch(1:2, :);
      
      rXY2 = sum(pXY.^2)';
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Hit Outer Cylinder
      
      % prepare quadratic equation parts
      a = sum(vXY.^2)';
      b = sum(pXY .* vXY)';
      c = rXY2 - obj.outerRadius^2;
      
      % cylinder impact possible if
      %    b < 0    => approaching the z-axis
      %    c >  0   => outside the outer cylinder
      %    rtArg > 0 => path interescts infinite cylinder
      rtArg = b.^2 - c .* a;
      isHit = b < 0 & c > 0 & rtArg > 0;
      
      rt = -Inf(size(rtArg));
      rt(isHit) = real(sqrt(rtArg(isHit)));
      
      % solve
      tHitOuter = (-b - rt) ./ a;
      if any(tHitOuter < 0)
        error('Negative hit time')
      end
      
      % find first impact, if any
      zHitCyl = pos(3, :)' + tHitOuter .* vel(3, :)';
      tHitOuter(abs(zHitCyl) > halfLen) = Inf;

      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Hit Inner Cylinder
      
      % prepare quadratic equation parts
      c = rXY2 - obj.innerRadius^2;
      
      % cylinder impact possible if
      %    c < 0  => inside the inner cylinder
      %    c > 0 & b < 0  => outside the inner cylinder and moving inward
      %    rtArg > 0 => path interescts infinite cylinder
      rtArg = b.^2 - c .* a;
      isHit = (c < 0 | (c > 0 & b < 0)) & rtArg > 0;
      
      rt = Inf(size(rtArg));
      rt(isHit) = real(sqrt(rtArg(isHit)));
      
      % solve
      tHitInner = (-b + rt) ./ a;
      if any(tHitInner < 0)
        error('Negative hit time')
      end
      
      % find first impact, if any
      zHitCyl = pos(3, :)' + tHitInner .* vel(3, :)';
      tHitInner(abs(zHitCyl) > halfLen) = Inf;

      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Hit Disks
      tHitTop = -(pos(3, :) - halfLen) ./ vel(3, :);
      xyHitTop = pXY + vXY .* repmat(tHitTop, 2, 1);
      r2HitTop = sum(xyHitTop.^2)';
      tHitTop(r2HitTop > obj.outerRadius^2) = Inf;
      tHitTop(r2HitTop < obj.innerRadius^2) = Inf;
      tHitTop(vel(3, :) >= 0) = Inf;
      tHitTop(tHitTop < 0) = Inf;
      tHitTop = tHitTop';
      
      tHitBottom = -(pos(3, :) + halfLen) ./ vel(3, :);
      xyHitBottom = pXY + vXY .* repmat(tHitBottom, 2, 1);
      r2HitBottom = sum(xyHitBottom.^2)';
      tHitBottom(r2HitBottom > obj.outerRadius^2) = Inf;
      tHitBottom(r2HitBottom < obj.innerRadius^2) = Inf;
      tHitBottom(vel(3, :) <= 0) = Inf;
      tHitBottom(tHitBottom < 0) = Inf;
      tHitBottom = tHitBottom';
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Combine
      tHit = min([tHitOuter, tHitInner, tHitTop, tHitBottom], [], 2);      
    end
  end
end % classdef
