% Cylinder - a cylinder for ray tracing (see Shape)
%

classdef Cylinder < Shape
  properties(GetAccess = public, SetAccess = private)
    length = 1;
    radius = 1;
    radiusSquared = 1;
  end
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Public Methods
  methods
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Constructor and Configuration
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function obj = Cylinder(varargin)
      switch nargin
        case 0
          % do nothing
        case 1
          arg = varargin{1};
          if isscalar(arg) && isa(arg, 'Cylinder')
            obj.setTo(arg);
          elseif isnumeric(arg) && isvector(arg) && numel(arg) == 2
            obj.setLengthAndRadius(arg(1), arg(2));
          else
            error('Single argument must be a 2-vector or a Cylinder.')
          end
        case 2
          arg1 = varargin{1};
          arg2 = varargin{2};
          if isscalar(arg1) && isnumeric(arg1) && ...
              isscalar(arg2) && isnumeric(arg2)
            obj.setLengthAndRadius(arg1, arg2);
          else
            error('Argument must be a numeric scalar.')
          end
        otherwise
          error('Constructor requires 0, 1 or 2 arguments.')
      end
    end
    
    function obj = copy(sourceObj)
      obj = Cylinder(sourceObj);
    end
    
    function setTo(obj, sourceObj)
      Shape.setTo(obj, sourceObj);
      obj.length = sourceObj.length;
      obj.radius = sourceObj.radius;
      obj.radiusSquared = sourceObj.radiusSquared;
    end
    
    function setLengthAndRadius(obj, lenArg, radArg)
      obj.length = lenArg;
      obj.radius = radArg;
      obj.radiusSquared = radArg^2;
      obj.maxRadius = radArg;
    end
    
    function setTopPosition(obj, pos)
      pos = Vec3.returnSingleVector(pos) - [0; 0; obj.length / 2];
      obj.setPosition(pos);
    end
    
    function setBottomPosition(obj, pos)
      pos = Vec3.returnSingleVector(pos) + [0; 0; obj.length / 2];
      obj.setPosition(pos);
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Shape Properties
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function vol = volume(obj)
      vol = pi * obj.radiusSquared * obj.length;
    end
    function area = surfaceArea(obj)
      area = 2 * pi * obj.radius * (obj.radius + obj.length);
    end
    function area = crossSectionZ(obj)
      area = pi * obj.radiusSquared;
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Surface Properties
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function vNorm = surfaceNormal(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      posT = pos';
      halfLen = obj.length / 2;
      
      isOnTop = (halfLen - posT(:, 3)) < Shape.surfaceTolerance;
      isOnBottom = (posT(:, 3) + halfLen) < Shape.surfaceTolerance;
      isOnRound = ~(isOnTop | isOnBottom);
      
      vNorm = posT';
      vNorm(3, :) = 0;
      vNorm(1:2, isOnTop) = 0;
      vNorm(3, isOnTop) = 1;
      vNorm(1:2, isOnBottom) = 0;
      vNorm(3, isOnBottom) = -1;
      vNorm(:, isOnRound) = Vec3.normalize(vNorm(:, isOnRound));
      
      vNorm = obj.mLocalToGlobal * vNorm;
    end
    
    function cSurf = surfaceColor(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      halfLen = obj.length / 2;
      
      isOnTop = abs(halfLen - pos(3, :)) < Shape.surfaceTolerance;
      isOnBottom = abs(pos(3, :) + halfLen) < Shape.surfaceTolerance;
      isOnRound = ~(isOnTop | isOnBottom);
      
      cSurf = Vec3.copy(obj.surfColor, size(pos, 2));
      if any(isOnTop)
        cSurf(:, isOnTop) = Vec3.sum(cSurf(:, isOnTop) * 0.9, 0.1 * [0; 0; 1]);
      end
      if any(isOnBottom)
        cSurf(:, isOnBottom) = Vec3.sum(cSurf(:, isOnBottom) * 0.9, 0.1 * [0; 1; 0]);
      end
      if any(isOnRound)
        cSurf(:, isOnRound) = Vec3.sum(cSurf(:, isOnRound) * 0.9, 0.1 * [1; 0; 0]);
      end
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Inside and Outside
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function [isOn, isOnRound, isOnTop, isOnBottom] = isOnSurface(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      posT = pos';
      halfLen = obj.length / 2;
      
      % is it on the round part?
      radXY = sum(posT(:, 1:2).^2, 2);
      isOnRound = posT(:, 3) <= halfLen & posT(:, 3) >= -halfLen & ...
        abs(radXY - obj.radiusSquared) < Shape.surfaceTolerance;
      
      % is it on the top?
      isOnTop = abs(posT(:, 3) - halfLen) < Shape.surfaceTolerance;
      
      % is it on the bottom?
      isOnBottom = abs(posT(:, 3) + halfLen) < Shape.surfaceTolerance;
      
      % is it on the surface?
      isOn = isOnRound | isOnTop | isOnBottom;
    end
    function isIn = isInside(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      posT = pos';
      
      % is it in the cylinder radius and between the ends?
      halfLen = obj.length / 2;
      radXY = sum(posT(:, 1:2).^2, 2);
      isIn = radXY < obj.radiusSquared & ...
        posT(:, 3) < halfLen & posT(:, 3) > -halfLen;
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Collision Checking
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function tHit = hitSurface(obj, pos, vel, isLocal)
      isLocal = nargin > 3 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        [pos, vel] = obj.transformTo(pos, vel);
      end
      halfLen = obj.length / 2;
      
      % make XY only vectors
      [posMatch, velMatch] = Vec3.matchSize(pos, vel);
      pXY = posMatch(1:2, :);
      vXY = velMatch(1:2, :);
      
      rXY2 = sum(pXY.^2)';
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Hit Cylinder
      
      % prepare quadratic equation parts
      a = sum(vXY.^2)';
      b = sum(pXY .* vXY)';
      c = rXY2 - obj.radiusSquared;
      
      % precompute and eliminate imaginary results
      rtArg = b.^2 - c .* a;
      isHit = rtArg > 0;
      
      % solve
      rt = Inf(size(rtArg));
      rt(isHit) = real(sqrt(rtArg(isHit)));
      
      tPos = (-b + rt) ./ a;
      tNeg = (-b - rt) ./ a;
      
      % find first impact, if any
      isPosOk = tPos > 0;
      isNegOk = tNeg > 0;
      
      tHitCyl = Inf(size(tPos));
      tHitCyl(isPosOk) = tPos(isPosOk);
      tHitCyl(isNegOk) = tNeg(isNegOk);

      zHitCyl = pos(3, :)' + tHitCyl .* vel(3, :)';
      tHitCyl(abs(zHitCyl) > halfLen) = Inf;

      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Hit Disks
      tHitTop = -(pos(3, :) - halfLen) ./ vel(3, :);
      xyHitTop = pXY + vXY .* repmat(tHitTop, 2, 1);
      r2HitTop = sum(xyHitTop.^2)';
      tHitTop(r2HitTop > obj.radiusSquared) = Inf;
      tHitTop(tHitTop < 0) = Inf;
      tHitTop = tHitTop';
      
      tHitBottom = -(pos(3, :) + halfLen) ./ vel(3, :);
      xyHitBottom = pXY + vXY .* repmat(tHitBottom, 2, 1);
      r2HitBottom = sum(xyHitBottom.^2)';
      tHitBottom(r2HitBottom > obj.radiusSquared) = Inf;
      tHitBottom(tHitBottom < 0) = Inf;
      tHitBottom = tHitBottom';
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Combine
      tHit = min([tHitCyl, tHitTop, tHitBottom], [], 2);      
    end
    
    function tHit = hitFromInside(obj, pos, vel, isLocal)
      isLocal = nargin > 3 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        [pos, vel] = obj.transformTo(pos, vel);
      end
      halfLen = obj.length / 2;
      
      % make XY only vectors
      [posMatch, velMatch] = Vec3.matchSize(pos, vel);
      pXY = posMatch(1:2, :);
      vXY = velMatch(1:2, :);
      
      rXY2 = sum(pXY.^2)';
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Hit Cylinder
      
      % prepare quadratic equation parts
      a = sum(vXY.^2)';
      b = sum(pXY .* vXY)';
      c = rXY2 - obj.radiusSquared;
      
      % assume that rtArg > 0
      rtArg = b.^2 - c .* a;
      rt = real(sqrt(rtArg));
      
      % assume that b + rt > 0, so that only one tHit is possible
      tHitCyl = (-b + rt) ./ a;
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Hit Disks
      tHitTop = -(pos(3, :)' - halfLen) ./ vel(3, :)';
      tHitTop(vel(3, :) <= 0) = Inf;
      tHitBottom = -(pos(3, :)' + halfLen) ./ vel(3, :)';
      tHitBottom(vel(3, :) >= 0) = Inf;
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Combine
      tHit = min([tHitCyl, tHitTop, tHitBottom], [], 2);      
    end
    
    function tHit = hitFromOutside(obj, pos, vel, isLocal)
      isLocal = nargin > 3 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        [pos, vel] = obj.transformTo(pos, vel);
      end
      halfLen = obj.length / 2;
      
      % make XY only vectors
      [posMatch, velMatch] = Vec3.matchSize(pos, vel);
      pXY = posMatch(1:2, :);
      vXY = velMatch(1:2, :);
      
      rXY2 = sum(pXY.^2)';
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Hit Cylinder
      
      % prepare quadratic equation parts
      a = sum(vXY.^2)';
      b = sum(pXY .* vXY)';
      c = rXY2 - obj.radiusSquared;
      
      % cylinder impact possible if
      %    b < 0    => approaching the z-axis
      %    c >  0   => outside the cylinder
      %    rtArg > 0 => path interescts infinite cylinder
      rtArg = b.^2 - c .* a;
      isHit = b < 0 & c > 0 & rtArg > 0;
      
      rt = -Inf(size(rtArg));
      rt(isHit) = real(sqrt(rtArg(isHit)));
      
      % solve
      tHitCyl = (-b - rt) ./ a;
      if any(tHitCyl < 0)
        error('Negative hit time')
      end
      
      % find first impact, if any
      zHitCyl = pos(3, :)' + tHitCyl .* vel(3, :)';
      tHitCyl(abs(zHitCyl) > halfLen) = Inf;

      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Hit Disks
      tHitTop = -(pos(3, :) - halfLen) ./ vel(3, :);
      xyHitTop = pXY + vXY .* repmat(tHitTop, 2, 1);
      r2HitTop = sum(xyHitTop.^2)';
      tHitTop(r2HitTop > obj.radiusSquared) = Inf;
      tHitTop(vel(3, :) >= 0) = Inf;
      tHitTop(tHitTop < 0) = Inf;
      tHitTop = tHitTop';
      
      tHitBottom = -(pos(3, :) + halfLen) ./ vel(3, :);
      xyHitBottom = pXY + vXY .* repmat(tHitBottom, 2, 1);
      r2HitBottom = sum(xyHitBottom.^2)';
      tHitBottom(r2HitBottom > obj.radiusSquared) = Inf;
      tHitBottom(vel(3, :) <= 0) = Inf;
      tHitBottom(tHitBottom < 0) = Inf;
      tHitBottom = tHitBottom';
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Combine
      tHit = min([tHitCyl, tHitTop, tHitBottom], [], 2);      
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Time to impact on an right cylinder,
    % starting at x and moving with a velocity v.
    %
    % ===> this is for a single pos, vel vectors!
    % ===> x must be outside of the cylinder!
    %
    % The cylinder is at radius r, with its axis on the Z-axis.
    % The base of the cyclinder is in the XY-plane, the top
    % is at z = h.
    
    function [tHit, nHit] = hitFromOutsideSingleLocal(obj, pos, vel)
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % break vectors down for quick access
      x0 = pos(1);
      y0 = pos(2);
      z0 = pos(3);
      
      vx = vel(1);
      vy = vel(2);
      vz = vel(3);
      
      r = obj.radius;
      r2 = obj.radiusSquared;
      h = obj.length / 2;
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % start with some quick and simple checks
      
      % is the particle going the right way along each axis?
      if (z0 > h && vz > 0) || (z0 < -h && vz < 0) || ...
          (y0 > r && vy > 0) || (y0 < -r && vy < 0) || ...
          (x0 > r && vx > 0) || (x0 < -r && vx < 0)
        tHit = Inf;
        nHit = [];
        return
      end
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % look for impacts on the end discs
      
      % bottom end
      if z0 < -h && vz > 0
        tHit = -(h + z0) / vz;
        xHit = pos + tHit * vel;
        
        if sum(xHit(1:2) .* xHit(1:2)) < r2
          nHit = 2;
          return
        end
      end
      
      % top end
      if z0 > h && vz < 0
        tHit = (h - z0) / vz;
        xHit = pos + tHit * vel;
        
        if sum(xHit(1:2) .* xHit(1:2)) < r2
          nHit = 3;
          return
        end
      end
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % look for impacts on the cylindrical surface
      
      % quadratic equation ingredients
      a = vx .* vx + vy .* vy;
      b = x0 .* vx + y0 .* vy;
      c = x0 .* x0 + y0 .* y0 - r2;
      rtArg = b .* b - c .* a;
      
      % bail out if no cylinder impact
      %   these conditions are:
      %    b >= 0    => not approaching the z-axis
      %    c <  0    => already inside!
      %    rtArg < 0 => path does not interesct infinite cylinder
      if b >= 0 || c < 0 || rtArg < 0
        tHit = Inf;
        nHit = [];
        return
      end
      
      % find solution for impact on the infinite cylindrical surface
      %  only the negative solution can be valid from outside
      rt = real(sqrt(rtArg));
      tHit = (-b - rt) ./ a;
      nHit = 1;
      
      % impact z-coordinate
      zHit = tHit * vz + z0;
      
      % is this on the finite cylindrical surface?
      if zHit < -h || zHit > h
        tHit = Inf;
        nHit = [];
      end
    end
  end
end % classdef
