% ShapeUnion - a union of shapes for ray tracing (see Shape)
%

classdef ShapeUnion < Shape
  properties(GetAccess = public, SetAccess = private)
    shapeList = {};
  end
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Public Methods
  methods
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Constructor and Configuration
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function obj = ShapeUnion(varargin)
      switch nargin
        case 0
          % do nothing
        case 1
          arg = varargin{1};
          if isscalar(arg) && isa(arg, 'ShapeUnion')
            obj.setTo(arg);
          elseif isa(arg, 'Shape')
            obj.shapeList = {arg};
          else
            error('Single argument must be a Shape or ShapeUnion object.')
          end
        otherwise
          for n = 1:numel(varargin)
            if ~isa(varargin{n}, 'Shape')
              error('Multiple arguments must be Shape objects.')
            end
          end
          obj.shapeList = varargin;
      end
    end
    
    function obj = copy(sourceObj)
      obj = ShapeUnion(sourceObj);
    end
    
    function setTo(obj, sourceObj)
      Shape.setTo(obj, sourceObj);
      obj.shapeList = sourceObj.shapeList;
      for n = 1:nargin
        obj.shapeList{n} = obj.shapeList{n}.copy;
      end
    end
    
    function setShapeList(obj, varargin)
      for n = 1:numel(varargin)
        if ~isa(varargin{n}, 'Shape')
          error('All arguments must be Shape objects.')
        end
      end
      obj.shapeList = varargin;
    end
      
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Surface Properties
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function vNorm = surfaceNormal(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      
      % find first shape which has pos on its surface
      [isOn, isOnShape] = isOnSurface(obj, pos, true);

      % compute surface normals
      vNorm = zeros(3, size(pos, 2));
      for n = 1:numel(obj.shapeList)
        if any(isOnShape(:, n))
          vNorm(:, isOnShape(:, n)) = ...
            obj.shapeList{n}.surfaceNormal(pos(:, isOnShape(:, n)));
        end
      end
      
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Inside and Outside
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function [isOn, isOnShape, isInShape] = isOnSurface(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
      
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      
      % find shapes which contain pos
      [isIn, isInShape] = obj.isInside(pos, true);
      
      % find shapes which have pos on their surface
      Nshape = numel(obj.shapeList);
      isOnShape = false(size(pos, 2), Nshape);
      for n = 1:Nshape
        isOnShape(:, n) = obj.shapeList{n}.isOnSurface(pos);
      end
%       fprintf('isOnSurface %d\n', Nshape)
%       [isOnShape, isInShape]
      
      % remove parts that are no part of the union
      for n = 1:Nshape
        % it can only be on this if it is not in any other shapes
        %   the surfaces of other shapes are excluded to avoid
        %   numerical problems
        nNotThis = [1:(n - 1), (n + 1):Nshape];
        isOnShape(:, n) = isOnShape(:, n) & ...
          ~any(isInShape(:, nNotThis) & ~isOnShape(:, nNotThis), 2);
      end
      isOn = any(isOnShape, 2);
    end
    
    function [isIn, isInShape] = isInside(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
      
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      
      % find all shapes which contain pos
      Nshape = numel(obj.shapeList);
      isInShape = false(size(pos, 2), Nshape);
      for n = 1:Nshape
        isInShape(:, n) = obj.shapeList{n}.isInside(pos);
      end
      isIn = any(isInShape, 2);
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Collision Checking
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function [tHit, tHitShape] = hitSurface(obj, pos, vel, isLocal)
      isLocal = nargin > 3 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        [pos, vel] = obj.transformTo(pos, vel);
      end
      
      % expand pos or vel as needed
      [pos, vel, Nvec] = Vec3.matchSize(pos, vel);
      
      % compute time to hit each shape
      Nshape = numel(obj.shapeList);
      tHitShape = zeros(Nvec, Nshape);
      
      %%%%% WRONG - if the first hit fails, you have to find the next!
      warning('function incomplete')
      
      for n = 1:Nshape
        % compute time to hit
        tHitShape(:, n) = obj.shapeList{n}.hitSurface(pos, vel);
        
        % compute hit positions
        isHit = isfinite(tHitShape(:, n));
        if any(isHit)
          posHit = Shape.hitPosition(pos(:, isHit), ...
            vel(:, isHit), tHitShape(isHit, n));
        
          % is this point on the remaining surface?
          [isOn, isOnShape] = obj.isOnSurface(posHit, true);
      
          isHit(isHit) = isHit(isHit) & isOnShape(:, n);
          tHitShape(~isHit, n) = Inf;
        end
      end
      
      % find first hit
      tHit = min(tHitShape, [], 2);
    end
    
    function tHit = hitFromInside(obj, pos, vel, isLocal)
      isLocal = nargin > 3 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        [pos, vel] = obj.transformTo(pos, vel);
      end
      
      % expand pos or vel as needed
      [pos, vel, Nvec] = Vec3.matchSize(pos, vel);
      
      % compute time to hit each shape
      Nshape = numel(obj.shapeList);
      tHitShape = zeros(Nvec, Nshape);
      
      %%%%% WRONG - if the first hit fails, you have to find the next!
      warning('function incomplete')

      for n = 1:Nshape
        % compute time to hit
        tHitShape(:, n) = obj.shapeList{n}.hitSurface(pos, vel);
        
        % compute hit positions
        isHit = isfinite(tHitShape(:, n));
        if any(isHit)
          posHit = Shape.hitPosition(pos(:, isHit), ...
            vel(:, isHit), tHitShape(isHit, n));
        
          % is this point on the remaining surface?
          [isOn, isOnShape] = obj.isOnSurface(posHit, true);
          
%           [tHitShape(isHit, :), isOnShape]
      
          isHit(isHit) = isHit(isHit) & isOnShape(:, n);
          tHitShape(~isHit, n) = Inf;
        end
      end
      
      % find first hit
      tHit = min(tHitShape, [], 2);
    end
    
    function tHit = hitFromOutside(obj, pos, vel, isLocal)
      isLocal = nargin > 3 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        [pos, vel] = obj.transformTo(pos, vel);
      end
      
      % expand pos or vel as needed
      [pos, vel, Nvec] = Vec3.matchSize(pos, vel);
      
      % compute time to hit each shape
      Nshape = numel(obj.shapeList);
      tHitShape = zeros(Nvec, Nshape);
      
      for n = 1:Nshape
        % compute time to hit
        tHitShape(:, n) = obj.shapeList{n}.hitFromOutside(pos, vel);
      end
      
      % find first hit
      tHit = min(tHitShape, [], 2);
    end
  end % methods
end % classdef
