% rslt = gasBox(p, isDraw)
%
%  Follow a gas molecule around a box to determine some statistical
%  distributions.  The box is assumed to be much smaller than the
%  particle mean-free-path, so collisions are ignored.
%
% The parameter struct (see getGasBoxParams):
% p.Ntrial = 1e3;
% p.tEnd = 1;
% p.dtBin = 1e-4;
% p.vMax = 4e3;
% p.dvBin = 1;
% 
% p.temp = 293;  % temperature, K
% p.mass = 2;    % particle mass, AMU
% 
% p.shapes = {...}; % list of shapes, first is the bounding box, second TM
%
% Usefull numbers: at 1e-8 mbar, there are 1e8 molecules / cm^3
%   and the mean-free-path is 10km.


function rslt = gasBox(p, isDraw)
  
  if nargin < 2
    isDraw = 0;
  end
  
  %%%% extract paramters
  struct2workspace(p);
  
  Nshape = numel(shapes);  
  
  ATOMIC_MASS = 1.66053886e-27;           % kg/au
  forceCoef = mass * ATOMIC_MASS / dtBin;

  % initialize histogram struct
  numberTimeBins = ceil(tEnd / dtBin);
  numberVelBins = ceil(vMax / dvBin);
  
  s0 = struct('numberTimeBins', numberTimeBins, ...
    'numberVelBins', numberVelBins, ...
    'forceVsTime', zeros(3, numberTimeBins), ...
    'velNormDist', zeros(1, numberVelBins));

  % make one for each surface
  rslt = repmat(s0, Nshape, 1);
  
  % clear for drawing
  if isDraw
    clf
  end
  
  %%%%%%%%%%%%
  % generate an initial position and launch velocity
  pos = getStartPos(shapes, Ntrial);
  vel = getMaxBoltzVel(temp, mass, Ntrial);
  
  %%%%%%%%%%%%
  % start at time zero
  t = zeros(Ntrial, 1);
    
  % initilize some loop variables
  vNormHit = zeros(size(vel));
  nStuck = zeros(Ntrial, 1);
  
  % error counters
  Nerr.stuck = 0;
  Nerr.escaped = 0;
  Nerr.notNormal = 0;
  
  % loop for specified time
  Ntodo = Ntrial;
  NtodoMax = floor(Ntodo / 20);  
  while Ntodo > NtodoMax
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % find time to next hit
    dtHit = zeros(Ntrial, Nshape);
    for n = 1:Nshape
      if n == 1
        % inside surface of the bounding box
        dtHit(:, n) = shapes{n}.hitFromInside(pos, vel);
      else
        % outside surface of a part
        dtHit(:, n) = shapes{n}.hitFromOutside(pos, vel);
      end
    end
    [dtMin, nMin] = min(dtHit, [], 2);
    
    % check for stuck particles
    isStuck = dtMin < 1e-9;
    nStuck(isStuck) = nStuck(isStuck) + 1;
    nStuck(~isStuck) = 0;
    isBad = nStuck > 2;
    if any(isBad)
%       [nStuck(isBad), dtMin(isBad), nMin(isBad)]
%       pos(:, isBad)
%       vel(:, isBad)
%       posHit(:, isBad)
%       vNormHit(:, isBad)
%       fprintf('Stuck particle! %g\n', min(t))
      
      % count stuck particles
      Nerr.stuck = Nerr.stuck + sum(isBad);
    end
    
    % check for escaped particles
    isEsc = ~isfinite(dtMin);
    if any(isEsc)
      Nerr.escaped = Nerr.escaped + sum(isEsc);
      isBad = isBad | isEsc;      
    end
    
    % prevent bad particles from having problems...
    dtMin(isBad) = 0;
    nMin(isBad) = 0;
      
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % hit time and position
    tHit = t + dtMin;
    posHit = Shape.hitPosition(pos, vel, dtMin);
      
    % normal vector at hit
    for n = 1:Nshape
      nHitThis = (nMin == n);
      if any(nHitThis)
        % surface normal
        if n == 1
          % inside surface of the bounding box
          vNormHit(:, nHitThis) = -shapes{n}.surfaceNormal(posHit(:, nHitThis));
        else
          % outside surface of a part
          vNormHit(:, nHitThis) = shapes{n}.surfaceNormal(posHit(:, nHitThis));
        end
      end
    end
    
    % next launch velcities
    velLaunch = getFluxVel(temp, mass, vNormHit, Ntrial);
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % record these events
    
    % normal velocity
    velNorm = -Vec3.dot(vNormHit, vel);
    isNegNorm = (velNorm < -1) & ~isBad;
    if any(isNegNorm)
%       [velNorm(isBad), dtMin(isBad), nMin(isBad)]
%       pos(:, isBad)
%       vel(:, isBad)
%       posHit(:, isBad)
%       vNormHit(:, isBad)
%       fprintf('Negative normal velocity %g\n', min(t))
      Nerr.notNormal = Nerr.notNormal + sum(isNegNorm);
    end
    
    % time and velocity bin index
    nTime = min(numberTimeBins, floor(tHit / dtBin) + 1);
    nVel = min(numberVelBins, floor(abs(velNorm) / dvBin) + 1);
    
    for n = 1:Nshape
      nHitThis = (nMin == n);
      if any(nHitThis)
        nn = find(nHitThis);
        nnTime = nTime(nn);
        nnVel = nVel(nn);
                
        % time series of impact force
        rslt(n).forceVsTime(:, nnTime) = rslt(n).forceVsTime(:, nnTime) + ...
          forceCoef * (vel(:, nHitThis) - velLaunch(:, nHitThis));
      
        % distribution of impact velocities
        rslt(n).velNormDist(nnVel) = rslt(n).velNormDist(nnVel) + 1;
      end
    end
 
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % draw stuff
    if isDraw == 1
      % draw all particles
      plot3(posHit(1, :)', posHit(2, :)', posHit(3, :)', 'r.')
      hold on
      line([pos(1, :); posHit(1, :)], [pos(2, :); posHit(2, :)], ...
        [pos(3, :); posHit(3, :)], 'Color', 'b')
      grid on
    elseif isDraw == 2
      % draw some particles
      isGap = abs(posHit(1, :)) < 0.2 & abs(posHit(2, :)) < 0.2 & ...
        posHit(3, :) < 0.01 & posHit(3, :) > -0.03;
      if any(isGap)
        plot3(posHit(1, isGap)', posHit(2, isGap)', posHit(3, isGap)', 'r.')
        hold on
        line([pos(1, isGap); posHit(1, isGap)], ...
          [pos(2, isGap); posHit(2, isGap)], ...
          [pos(3, isGap); posHit(3, isGap)], 'Color', 'b')
        grid on
        view(2)
      end
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % update particle trajectories
    isOk = ~isBad;
    t(isOk) = tHit(isOk);
    pos(:, isOk) = posHit(:, isOk);
    vel(:, isOk) = velLaunch(:, isOk);
    
    % restart bad particles
    Nbad = sum(isBad);
    pos(:, isBad) = getStartPos(shapes, Nbad);
    vel(:, isBad) = getMaxBoltzVel(temp, mass, Nbad);
      
    % update while condition
    isNotDone = t < tEnd;
    Ntodo = sum(isNotDone);
  end
  
  if isDraw
    % draw cylinders
    for n = 1:Nshape
      if isa(shapes{n}, 'Cylinder')
        tmShape = shapes{n};
        tmDiam = tmShape.radius * 2;
        tmThick = tmShape.length;
        tmDir = tmShape.orientation;
        tmPos = tmShape.position + tmThick * tmDir / 2;
        drawMirror(tmDiam, tmThick, tmPos, tmDir, [0; 1; 0], [1; 0; 0]);
      end
    end
  end
  
  % output error counters
  Nt = tEnd * Ntrial;
  if (Nerr.stuck > Nt) || (Nerr.escaped > 0) || (Nerr.notNormal > 0)
    fprintf('  Nstuck = %d, Nescaped = %d, NnotNormal = %d\n', ...
      Nerr.stuck, Nerr.escaped, Nerr.notNormal);
  end
end

  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% is this position in any of the listed shapes?
function pos = getStartPos(shapes, N)
  bndBox = shapes{1};
  boxSize = bndBox.length / 2;
  
  pos = rand(3, N) * max(boxSize);
  
  isInvalid = ~bndBox.isInside(pos) | isInAnyShape(shapes(2:end), pos) ;
  while any(isInvalid)
    % make some new positions
    pos(:, isInvalid) = rand(3, sum(isInvalid)) * max(boxSize);
    
    isInvalid(isInvalid) = ~bndBox.isInside(pos(:, isInvalid)) | ...
      isInAnyShape(shapes(2:end), pos(:, isInvalid)) ;
  end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% is this position in any of the listed shapes?
function isInAny = isInAnyShape(shapes, pos)
  isInAny = false(numel(pos, 2), 1);
  for n = 1:numel(shapes)
    isInAny = isInAny | shapes{n}.isInside(pos);
  end
end
